#NoTrayIcon
#Region ;**** Directives created by AutoIt3Wrapper_GUI ****
#AutoIt3Wrapper_Outfile=Chip8_x32.exe
#AutoIt3Wrapper_Outfile_x64=Chip8_x64.exe
#AutoIt3Wrapper_Compile_Both=y
#AutoIt3Wrapper_UseX64=y
#EndRegion ;**** Directives created by AutoIt3Wrapper_GUI ****
;Written along this youtube series: https://www.youtube.com/watch?v=AsukaPLuTsU
;Bugfixing along with the https://www.youtube.com/watch?v=YvZ3LGaNiS0
;Converted then from BlitzBasic to AutoIt in December 2024
#include <GUIConstantsEx.au3>
#include <FontConstants.au3>
#include <WindowsConstants.au3>
#include <FileConstants.au3>
#include <Misc.au3>
#include <APIMiscConstants.au3>
#include <WinAPIMisc.au3>
#include <Sound.au3>
#include <ComboConstants.au3>
#include <GuiComboBox.au3>
#include <EditConstants.au3>
#include <GuiEdit.au3>

Global $pWav, $tWav, $bPlay = False, $MuteSound = 0
$tmp = ""
If StringRight(@ScriptDir, 1) <> "\" Then $tmp = "\"
Global $iniFile = @ScriptDir & $tmp & "Chip8.ini"


Opt("TrayAutoPause", 0)
Opt("GUIResizeMode", BitOR($GUI_DOCKSIZE, $Gui_DOCKLEFT))
Opt("GUICloseOnESC", 0)

;Gui Buttons:
Global $BU_PP, $BU_NP, $BU_F, $BU_P, $BU_B, $BU_PD, $BU_D, $BU_U, $BU_PU, $BU_T, $BU_PC
Global $Bu_Start, $Bu_Step, $Bu_SpeedM, $Bu_SpeedP, $Bu_BGC, $Bu_FGC, $Bu_Reset, $Bu_Reboot, $Bu_Load

;Chip8 handling
Global $totalmemory = 4096
Global $Display[(64 * 32) + 1]
Global $mem[$totalmemory]
Global $MemCopy[$totalmemory]
Global $Stack[16]
Global $V[16]
Global $keys[16]
Global $Chip8Version = 1          ;1 = Chip8
Global $ChipQuirkFix = 0		  ;some games may work with this on
Global $I, $Pc, $StackPointer, $DelayTimer, $SoundTimer, $Stimer = TimerInit(), $Dtimer = TimerInit()
Global $needRedraw = False
Global $anykeypressed = -2
Global $step = 0
Global $FontStartAddress = 0x50, $go = 1, $cycle
Global $PokeStart = 0, $RomName = ""
Global $PathToSoundFile
Global $DisplayLabel
Global $PixNr = 1
Global $Pix = ChangePixelType($PixNr)      ;Pixel display as $pix
Global $sounddriver = 1
Global $monitorValues[10]
Global $memGraphDump[60][5]
Global $loopdetect_count = 0, $loopdetect_oldpc = 0, $loopdetect_PC = 0


Global $KSN = 0, $KeySets[10][7] , $KCC[7],  $KeyConfigForm, $NrComboBoxPreset, $KeyConfigInfoTxt   ;KSN=Key set number   KCC = KeyConfig combo boxes
Global $keyNr[] = [58, 31, 32, 33, 51, 57, 45, 41, 53, 44, 59, 43, 34, 52, 46, 56]
;Key Meanings     " x    1  2   3   q   w   e   a    s   d   y   c   4    r  f   v"

Global $KeySets[10][7], $KSN = 1
Global Const $KeyN = "x123qweasdyc4rfv"

ReadCustomKeys()

Global $Colors[] = [0xffffff, 0x000000]

Global $sprite[17]                        ;Used to draw the current sprite to the monitor

;Setup the default sound file
$tmp = ""
If StringRight(@ScriptDir, 1) <> "\" Then $tmp = "\"
$PathToSoundFile = @ScriptDir & $tmp & "beep.wav"

;Create a default key for sound
If IniRead($iniFile, "config", "SoundFile", -1) = -1 Then IniWrite($iniFile, "config", "SoundFile", $PathToSoundFile)
If IniRead($iniFile, "config", "SoundDriver", -1) = -1 Then
	IniWrite($iniFile, "config", "SoundDriver", $sounddriver)
	IniWrite($iniFile, "config", "SoundDriverInfo", ";0=WinApi 1=_SoundPlay")
EndIf

If $sounddriver < 1 Or $sounddriver > 2 Then $sounddriver = 1

;Read config from the Ini file:
$Colors[0] = IniRead($iniFile, "config", "ColorFG", $Colors[0])
$Colors[1] = IniRead($iniFile, "config", "ColorBG", $Colors[1])
$MuteSound = IniRead($iniFile, "config", "MuteSound", $MuteSound)
$PathToSoundFile = IniRead($iniFile, "config", "SoundFile", $PathToSoundFile)
$sounddriver = IniRead($iniFile, "config", "SoundDriver", $sounddriver)


$Gui_hWnd = GUICreate("Chip 8:", 715, 244, -1, -1, -1, $WS_EX_COMPOSITED)

$BU_PP = GUICtrlCreateButton("<Page", 333, 222, 40, 20)
$BU_NP = GUICtrlCreateButton("Page>", 376, 222, 40, 20)
$BU_F = GUICtrlCreateButton("F", 449, 222, 20, 20)
$BU_P = GUICtrlCreateButton("P", 470, 222, 20, 20)
$BU_PC = GUICtrlCreateButton("PC", 490, 222, 20, 20)
$BU_B = GUICtrlCreateButton("B", 525, 222, 20, 20)
$BU_PD = GUICtrlCreateButton("PD", 556, 222, 30, 20)
$BU_D = GUICtrlCreateButton("D", 595, 222, 25, 20)
$BU_U = GUICtrlCreateButton("U", 621, 222, 25, 20)
$BU_PU = GUICtrlCreateButton("PU", 656, 222, 30, 20)
$BU_T = GUICtrlCreateButton("T", 695, 222, 20, 20)


GUICtrlSetTip($BU_PP, "Previous page", "Info")
GUICtrlSetTip($BU_NP, "Next page", "Info")
GUICtrlSetTip($BU_F, "On Monitor page: Jump to Font Start" & @CRLF & "Else: Change to font 1", "Multi Button")
GUICtrlSetTip($BU_P, "On Monitor page: Jump to Program start" & @CRLF & "Else: Change to font 2", "Multi Button")
GUICtrlSetTip($BU_PC, "On Monitor page: Jump to Program Counter" & @CRLF & "Else: Change the pixel type of the Display", "Multi Button")
GUICtrlSetTip($BU_B, "On Monitor page: Jump to End" & @CRLF & "Else: Open the Activity log window" & @CRLF & "or moves the activity window to the main window", "Multi Button")
GUICtrlSetTip($BU_PD, "On Monitor page: Page Down" & @CRLF & "Else: Open the extra Keys Config Window" , "Multi Button")
GUICtrlSetTip($BU_D, "Line Down", "Monitor")
GUICtrlSetTip($BU_U, "Line Up", "Monitor")
GUICtrlSetTip($BU_PU, "Page Up", "Monitor")
GUICtrlSetTip($BU_T, "On Monitor page: Jump to 0" & @CRLF & "Else: Open the ini file in the Editor", "Multi Button")


$Bu_Start = GUICtrlCreateButton("Run", 1, 222, 35, 20)
$Bu_Step = GUICtrlCreateButton("Step", 37, 222, 35, 20)
$Bu_SpeedM = GUICtrlCreateButton("-", 74, 222, 22, 20)
$Bu_SpeedP = GUICtrlCreateButton("+", 96, 222, 22, 20)
$Bu_BGC = GUICtrlCreateButton("BC", 122, 222, 26, 20)
$Bu_FGC = GUICtrlCreateButton("FC", 148, 222, 26, 20)
$Bu_Reset = GUICtrlCreateButton("R", 177, 222, 26, 20)
$Bu_Reboot = GUICtrlCreateButton("B", 203, 222, 26, 20)
$Bu_Mute = GUICtrlCreateButton("M", 232, 222, 26, 20)
$Bu_Chip8 = GUICtrlCreateButton("8", 260, 222, 13, 20)
$Bu_QuirkFix = GUICtrlCreateButton("9", 274, 222, 13, 20)
$Bu_Load = GUICtrlCreateButton("Load", 287, 222, 41, 20)

GUICtrlSetTip($Bu_Start, "Start/Stop", "Emulator")
GUICtrlSetTip($Bu_Step, "Stepping mode, when the emulator is stopped.", "Emulator")
GUICtrlSetTip($Bu_SpeedM, "Decrease Speed delay Value (faster)", "Emulator")
GUICtrlSetTip($Bu_SpeedP, "Increase Speed delay Value (Slower)", "Emulator")
GUICtrlSetTip($Bu_BGC, "Change Background Color", "Display")
GUICtrlSetTip($Bu_FGC, "Change Foreground Color", "Display")
GUICtrlSetTip($Bu_Reset, "Reset and Stop the emulation", "Emulator")
GUICtrlSetTip($Bu_Reboot, "Reboot the Emulation. Reloads one of the Built-In Roms", "Emulator")
GUICtrlSetTip($Bu_Mute, "Mute Sound", "Emulator")
GUICtrlSetTip($Bu_Chip8, "Fix for some Chip8 only settings" & @CRLF & "Use it if there are Game glitches", "Emulator")
GUICtrlSetTip($Bu_QuirkFix, "Another option that may fix some games" & @CRLF & "Hidden and Blinky work if this turned on", "Emulator")
GUICtrlSetTip($Bu_Load, "Load an (.ch8) file into emulators memory", "Emulator")

GUICtrlSetBkColor($Bu_Reboot, 0xff0000)
GUICtrlSetColor($Bu_Reboot, 0xffffff)

GUICtrlSetBkColor($Bu_Reset, 0xff0000)
GUICtrlSetColor($Bu_Reset, 0xffffff)


Global $DisplayLabel = GUICtrlCreateLabel("", 2, 2, 325, 194)
GUICtrlSetBkColor($DisplayLabel, $Colors[1])
GUICtrlSetColor($DisplayLabel, $Colors[0])
GUICtrlSetFont($DisplayLabel, 5, $FW_BOLD, $GUI_FONTNORMAL, "Terminal", $DRAFT_QUALITY)
GUISetBkColor(0x000000)

Global $Info_Label = GUICtrlCreateLabel("", 330, 0, 383, 222)
GUICtrlSetBkColor($Info_Label, 0x5040FF)
GUICtrlSetColor($Info_Label, 0xffffff)
SetInfoFont(0)

Global $input = GUICtrlCreateInput("Test", 2, 198, 325, 22)
GUICtrlSetFont($input, 9, $FW_Normal, $GUI_FONTNORMAL, "Segoe UI", $DRAFT_QUALITY)

Global $infotimer = TimerInit()
Local $InfoNr = 0
Global $infoChanged = 0
Global $infoMaX = 5                ;Maximum number of information pages
Global $MemTopAddr = 0, $MemDisplay = 16, $memLines = 15 ; Memory top address = Where to start looking in the array. $memdisplay = Collumns. $memlines=Rows (starting from 0 to 10 = 11 lines)
Global $memstop = Int(($totalmemory - 1) / ($MemDisplay * $memLines)) * ($MemDisplay * $memLines) ;Calculate where the memory monitor should stop
Local $info
$info = "(1/" & $infoMaX + 1 & ")" & " Keyboard Bindings:" & @CRLF & @CRLF & "   Chip 8    This Emulator" & @CRLF & @CRLF & "    123C         1234" & @CRLF & "    456D         QWER" & @CRLF & "    789E         ASDF" & @CRLF & "    A0BF         YXCV"

GUICtrlSetData($Info_Label, $info)
GUISetState()

Global $dbggui = GUICreate("Chip 8 Activity Log (before calculation)", 530, 482, -1, -1, 0, 0)
Global $dbgedit = GUICtrlCreateEdit("", 0, 5, 525, 395)
Global $dbgeditCLS = GUICtrlCreateButton("CLS", 475, 405, 44, 19)
Global $dbgeditHide = GUICtrlCreateButton(" Hide ", 475, 429, 44, 19)
Global $dbgeditRun = GUICtrlCreateButton(" Run ", 344, 429, 44, 19)
Global $dbgeditUSE = GUICtrlCreateButton("Use", 370, 407, 26, 15)
;DllCall("user32.dll", "lresult", "SendMessageW", "hwnd", $dbgedit, "uint", $EM_SETLIMITTEXT, "wparam", 100000000, "lparam" , 0) ; Used from _GuiCtrlEdit_SetLimitText
_GUICtrlEdit_SetLimitText($dbgedit, 100000000)
GUICtrlSetFont($dbgedit, 13, $FW_Normal, $GUI_FONTNORMAL, "Fixedsys", $DRAFT_QUALITY)
$dbgeditInput1 = GUICtrlCreateInput("", 3, 429, 91, 21, $ES_NUMBER)
$dbgeditRPC = GUICtrlCreateRadio("PC", 3, 400, 46, 23)
$dbgeditRCY = GUICtrlCreateRadio("Cycle", 49, 400, 58, 23)
$dbgeditRVR = GUICtrlCreateRadio("Vreg", 116, 400, 56, 23)
$dbgeditRIn = GUICtrlCreateRadio("Instruction", 190, 400, 86, 23)
$dbgeditRI = GUICtrlCreateRadio("I", 280, 400, 20, 23)
$dbgeditRST = GUICtrlCreateRadio("Stack", 320, 400, 45, 23)
$dbgeditCoVr = GUICtrlCreateCombo("", 116, 429, 62, 25, BitOR($ES_READONLY, $CBS_DROPDOWNLIST))
$dbgeditCoIn = GUICtrlCreateCombo("", 190, 429, 72, 25, BitOR($ES_READONLY, $CBS_DROPDOWNLIST, $WS_VSCROLL, $CBS_DISABLENOSCROLL))
$dbgeditCheckActive = GUICtrlCreateCheckbox("Active", 280, 426, 52, 26)
$dbgeditInputDec = GUICtrlCreateInput(0, 402, 405, 61, 21, $ES_NUMBER)
$dbgeditInputHex = GUICtrlCreateInput(0, 402, 429, 61, 21)
GUICtrlSetTip($dbgeditRPC, "Enter a number below (Decimal) (0 to 4095)")
GUICtrlSetTip($dbgeditRCY, "Enter a number below (Decimal)", "Stop at Cycle #")
GUICtrlSetTip($dbgeditRIn, "Select a number from the box below this", "Stop and Instruction #")
GUICtrlSetTip($dbgeditRVR, "Select a number from the box below this to stop the emulator" & @CRLF & "when that V register contains a number from the inputbox to the left" & @CRLF & "Numbers in range from 0 to 255", "Stop at Vreg")
GUICtrlSetTip($dbgeditRI, "Stop the emulator" & @CRLF & "when the I pointer contains a number from the inputbox to the far left (Decimal)" & @CRLF & "Numbers in range from 0 to 4095", "Stop at I")
GUICtrlSetTip($dbgeditRST, "Stop the emulator" & @CRLF & "when the (latest) Stack contains a number from the inputbox to the far left (Decimal)" & @CRLF & "Numbers in range from 0 to 4095", "Stop at Stack value")
GUICtrlSetTip($dbgeditCheckActive, "Allows the emulation to be stopped at the selected options")
GUICtrlSetTip($dbgeditInputDec, "Enter a number (Decimal)")
GUICtrlSetTip($dbgeditInputHex, "Enter a number (HexaDecimal) (0123456789ABCDEF)")
GUICtrlSetTip($dbgeditHide, "Stop debugging and close the window")
GUICtrlSetTip($dbgeditRun, "Start/Stop the emulation")
GUICtrlSetTip($dbgeditCLS, "Clear the Output")
GUICtrlSetTip($dbgeditInput1, "Enter a number which will be used to stop the emulation")
GUICtrlSetTip($dbgeditUSE, "Use this decimal number as a search value" & @CRLF & "Copies the number over to the input box")
GUICtrlSetState($dbgeditRPC, $GUI_CHECKED)
GUICtrlSetLimit($dbgeditInputHex, 7)
GUICtrlSetLimit($dbgeditInputDec, 8)
$tmp = ""
For $x = 0 To 15
	$tmp = $tmp & $x & "|"
Next
GUICtrlSetData($dbgeditCoVr, $tmp, 0)

$tmp = "00E0|00EE|1000|2000|3000|4000|5000|6000|7000|8000|8001|8002|8003|8004|8005|8006|8007|800E|9000|A000|B000|C000|D000|E09E|E0A1|F007|F00A|F015|F018|F01E|F029|F033|F055|F065"
GUICtrlSetData($dbgeditCoIn, $tmp, "1000")


GUISwitch($Gui_hWnd)

Global $DBGOn = 0, $DBGWinOn = 0

Init()
AddDefaultFont()
AddTestRom()
Global $hDLL = DllOpen("user32.dll")        ;_isPressed

$fname = @ScriptDir
Global $framecounter = TimerInit()
$lastInput = ""
$speed = 16

Global $edittext

WinSetTitle($Gui_hWnd, "", "Chip 8 - " & $RomName)

LoadSound($PathToSoundFile)
UpdateInfo()

KeyConfig()

Global $iStopTest = 0
GUIRegisterMsg($WM_COMMAND, "MY_WM_COMMAND")
GUIRegisterMsg($WM_SYSCOMMAND, "MY_SYS_COMMAND")

While 1

	If $DBGOn = 1 Then
		If $DBGWinOn = 0 Then
			$DBGWinOn = 1
			GUISetState(@SW_SHOW, $dbggui)
		EndIf
	Else
		If $DBGWinOn = 1 Then
			$DBGWinOn = 0
			GUICtrlSetData($dbgedit, "")
			GUISetState(@SW_HIDE, $dbggui)
		EndIf
	EndIf

	$edittext = StringMid("StoppedRunning", ($go * 7) + 1, 7) & "   Chip8=(" & $Chip8Version & ") Fix (" & $ChipQuirkFix & ") Speed: (" & $speed & ")  SND: (" & (1 - $MuteSound) & ") (" & $cycle & ")"

	If $go = 1 Then
		If TimerDiff($framecounter) > $speed Then
			ChipRun()
			$framecounter = TimerInit()
		EndIf
	EndIf

	If $go = 0 And $step = 1 Then
		ChipRun()
		$step = 0
		$go = 0
	EndIf

	$lastInputTmp = GUICtrlRead($input)
	If $lastInput <> $lastInputTmp Then
		$lastInput = $lastInputTmp
	EndIf

	If $infoChanged = 1 Or TimerDiff($infotimer) > 200 Then
		GUICtrlSetData($input, $edittext)
		$tmpI = ""
		If $InfoNr < 0 Then $InfoNr = 0
		If $InfoNr > $infoMaX Then $InfoNr = $infoMaX
		If $InfoNr = 0 Then $tmpI = $info & @CRLF & $KeyConfigInfoTxt

		$tmpP = "(" & $InfoNr + 1 & "/" & $infoMaX + 1 & ")"

		If $InfoNr = 1 Then						;Displays the Currently drawn Sprite
			$tmpI = $tmpP & " Sprite Data:           " & $cycle & @CRLF
			For $y = 0 To $sprite[0]
				$tmp1 = "       "
				If $y = 2 Then $tmp1 = "PC= " & Hex($Pc, 3)
				If $y = 3 Then $tmp1 = "I = " & Hex($I, 3)
				$tmpI = $tmpI & $tmp1 & "             " & $sprite[$y + 1] & @CRLF
			Next
			$tmp1 = ""
		EndIf

		If $InfoNr = 2 Then						;Instruction/Register view
			For $x = 0 To 15
				;$tmpI = $tmpI & "V(" & $x & ")= " & Hex($V[$x], 2) & "  " & "V(" & $x + 8 & ")= " & Hex($V[$x + 8], 2) & @CRLF
				$tmpI = $tmpI & "V(" & Hex($x, 2) & ")=" & Hex($V[$x], 2) & "   " & "S(" & Hex($x, 1) & ")=" & Hex($Stack[$x], 3)
				If $x = 0 Then $tmpI = $tmpI & "  Cycle " & $cycle
				If $x = 1 Then $tmpI = $tmpI & "  PC:   " & Hex($Pc, 3) & "  " & $Pc
				If $x = 2 Then $tmpI = $tmpI & "  I=    " & Hex($I, 4) & " " & $I
				If $x = 3 Then $tmpI = $tmpI & "  SP=   " & $StackPointer
				$tmpI = $tmpI & @CRLF
			Next
			$tmpI = $tmpP & " V Register and Stack" & @CRLF & @CRLF & $tmpI
		EndIf

		If $InfoNr = 3 Then
			$tmpI = $tmpI & "Opcode:....." & Hex($monitorValues[0], 4) & "  (" & $monitorValues[0] & ")" & @CRLF
			$tmpI = $tmpI & "Instruction " & Hex($monitorValues[1], 4) & "  (" & $monitorValues[1] & ")" & @CRLF
			$tmpI = $tmpI & "Address:....." & Hex($monitorValues[2], 3) & "  (" & $monitorValues[2] & ")" & @CRLF
			$tmpI = $tmpI & "byte:........." & Hex($monitorValues[5], 2) & "  (" & $monitorValues[5] & ")" & @CRLF
			$tmpI = $tmpI & "Value.........." & Hex($monitorValues[6], 1) & "  (" & $monitorValues[6] & ")" & @CRLF
			$tmpI = $tmpI & "Vx:.........." & Hex($monitorValues[3], 1) & "    (" & $monitorValues[3] & ")" & @CRLF
			$tmpI = $tmpI & "Vy:..........." & Hex($monitorValues[4], 1) & "   (" & $monitorValues[4] & ")" & @CRLF & @CRLF
			$tmpI = $tmpI & "Delay= " & Hex($DelayTimer, 2) & "  Sound= " & Hex($SoundTimer, 2) & " Sound is " & StringMid("On Off", ($MuteSound * 3) + 1, 3) & @CRLF & @CRLF
			$tmpI = $tmpI & $monitorValues[9] & @CRLF
			$tmpI = $tmpI & $monitorValues[8] & @CRLF & @CRLF
			$tmpI = $tmpI & "ErrorMsg: " & $monitorValues[7]
			$tmpp1 = @CRLF & "PC: " & Hex($Pc, 3) & " cycle " & $cycle & "  I= " & Hex($I, 4)
			$tmpI = $tmpP & " Emulator state" & @CRLF & $tmpp1 & @CRLF & @CRLF & $tmpI
		EndIf

		If $InfoNr = 4 Then					;Hex view of the Chip8 memory
			$tmp = 0
			$tmp = $MemTopAddr

			$Text = $tmpP & " Monitor - Rom Data" & $Pc & ":" & $tmp & @CRLF & @CRLF
			For $y = 0 To $memLines
				;ConsoleWrite($tmp & "_" & Hex($tmp, 3) & @CRLF)

				$Text = $Text & (Hex(Int($tmp), 3)) & ": "                ;Bug Note the normal HEX($tmp,3) causes an error when the button $BU_PC is clicked (it sets $memtopaddress to $pc but the display of the Numbers NNN: shows 000 ..
				For $x = 0 To $MemDisplay - 1
					$b = -1
					If $tmp <= 0xfff Then $b = $mem[$tmp]
					$tmp = $tmp + 1
					If $b > -1 Then
						$Text = $Text & Hex($b, 2)
					Else
						$Text = $Text & "  "
					EndIf
				Next
				$Text = $Text & @CRLF
			Next
			$tmpI = $Text
			$Text = ""
			$tmp = ""
		EndIf

		If $InfoNr = 5 Then            ;Graphical display of font data
			$tmp = 0
			$tmp1 = ""
			$tmpNr = 0
			$tmpOffset = 0


			For $col = 0 To 16
				For $y = 0 To 4                         ;Make a Char from 5 bytes
					$tmp1 = $mem[$tmp + $FontStartAddress]
					$memGraphDump[$tmpNr][$y] = ""
					For $x = 3 To 0 Step -1                ;Make a line of pixels from 1 byte
						$tmppix = BitRead($tmp1, 5 + $x)
						If $tmppix = 1 Then
							$memGraphDump[$tmpNr][$y] = $memGraphDump[$tmpNr][$y] & $Pix
						Else
							$memGraphDump[$tmpNr][$y] = $memGraphDump[$tmpNr][$y] & " "
						EndIf
					Next
					$tmp = $tmp + 1
				Next
				$tmpNr = $tmpNr + 1
			Next

			For $col = 0 To 4
				For $y = 0 To 4
					For $x = 0 To 5
						$tmpI = $tmpI & " " & $memGraphDump[$x + ($col * 6)][($y)]
					Next
					$tmpI = $tmpI & @CRLF
				Next
				$tmpI = $tmpI & @CRLF
			Next
			$tmpI = $tmpP & " Fontset: " & @CRLF & $tmpI
			$tmp1 = ""
		EndIf


		GUICtrlSetData($Info_Label, $tmpI)
		$infoChanged = 0
		$infotimer = TimerInit()
		$tmpI = ""
	EndIf

	If $needRedraw = True Then DrawScreen()
WEnd

Func ChipRun()

	Local $opcode, $instruction, $address, $Vx, $Vy, $value, $byte, $carry

	$opcode = (BitShift($mem[$Pc], -8)) + $mem[$Pc + 1]
	$instruction = BitAND($opcode, 0xF000)
	$address = BitAND($opcode, 0x0FFF)
	$Vx = BitShift(BitAND($opcode, 0x0F00), 8)
	$Vy = BitShift(BitAND($opcode, 0x00F0), 4)
	$byte = BitAND($opcode, 0x00FF)
	$value = BitAND($opcode, 0x000F)
	$carry = 0
	$cycle = $cycle + 1

	$monitorValues[0] = $opcode
	$monitorValues[1] = $instruction
	$monitorValues[2] = $address
	$monitorValues[3] = $Vx
	$monitorValues[4] = $Vy
	$monitorValues[5] = $byte
	$monitorValues[6] = $value
	$monitorValues[7] = ""            ;Error messages
	$monitorValues[8] = ""
	$monitorValues[9] = ""
	GetTheKeys()

	If $DBGWinOn = 1 Then _GUICtrlEdit_AppendText($dbgedit, ($cycle & ": " & Hex($opcode, 4) & " " & Hex($instruction, 4) & " " & Hex($address, 3) & " " & Hex($Vx, 1) & " " & Hex($Vy, 1) & " " & Hex($byte, 2) & " " & Hex($value, 1) & " V(" & $Vx & ")=" & $V[$Vx] & " V(" & $Vy & ")=" & $V[$Vy] & " I=" & Hex($I, 3) & @CRLF))
	;ConsoleWrite (Hex($opcode,4) & " " & Hex($instruction,4) & " " & Hex($address,3) & " " & Hex($Vx,1) & " " & Hex($Vy,1) & " " & Hex($byte,2) & " " & Hex($value,1) & @CRLF)
	$loopdetect_oldpc = $Pc

	Switch $instruction
		Case 0x0000

			Switch $byte
				Case 0xE0 ; $E0 Clear Screen
					ClearDisplay()
					$Pc = $Pc + 2
					$needRedraw = True
					$monitorValues[9] = "00E0 Clear Display"
				Case 0xEE ;$EE Return from Subroutine
					$StackPointer = $StackPointer - 1
					If $StackPointer >= 0 And $StackPointer <= 15 Then
						$Pc = $Stack[$StackPointer] + 2
					Else
						$monitorValues[7] = "Stack error: Underflow"
						$go = 0
						If $bPlay = True Then SNDPlay(0)
					EndIf
					$monitorValues[8] = "Stackpointer=" & $StackPointer
					$monitorValues[9] = "00EE Return from Subroutine"
				Case Else  ; 0NNN Call RCA 1802 program at $address NNN
					$monitorValues[7] = "0NNN - Undefined"
					$go = 0
					If $bPlay = True Then SNDPlay(0)
			EndSwitch

		Case 0x1000    ;1NNN Jumps to $address NNN
			$Pc = $address
			$monitorValues[8] = "Address = " & Hex($address, 3)
			$monitorValues[9] = "1NNN Jump to address"


		Case 0x2000  ;2NNN Calls subroutine at NNN
			If $StackPointer >= 0 And $StackPointer <= 15 Then
				$Stack[$StackPointer] = $Pc
				$StackPointer = $StackPointer + 1
			Else
				$monitorValues[7] = "Stack error: Overflow"
				$go = 0
				If $bPlay = True Then SNDPlay(0)
			EndIf
			$Pc = $address
			$monitorValues[8] = "Address: " & $Pc & " SP+=" + $StackPointer
			$monitorValues[9] = "2NNN Call a subroutine"
		Case 0x3000    ;3XNN Skips the next $instruction if $Vx equals NN
			;If V[$Vx] = $byte Then $pc=$pc+2
			$Pc = $Pc + 2 + (($V[$Vx] = $byte) * 2)
			$monitorValues[8] = "Vx= " & $V[$Vx] & " byte=" & $byte & " Result:" & ($V[$Vx] = $byte)
			$monitorValues[9] = "3XNN Skip if Vx=Byte"
		Case 0x4000        ;4XNN Skip the next $instruction if $Vx not Equal to NN
			;If V[$Vx]<>$byte Then $pc=$pc+2
			$Pc = $Pc + 2 + (($V[$Vx] <> $byte) * 2)
			$monitorValues[8] = "Vx= " & $V[$Vx] & " byte=" & $byte & " Result:" & ($V[$Vx] <> $byte)
			$monitorValues[9] = "4XNN Skip if Vx<>Byte"
		Case 0x5000        ;5XY0 Skips the next $instruction if $Vx = VY
			;If V[$Vx] = V[Vy] Then $pc=$pc+2
			$Pc = $Pc + 2 + (($V[$Vx] = $V[$Vy]) * 2)
			$monitorValues[8] = "Vx= " & $V[$Vx] & " byte=" & $byte & " Result:" & ($V[$Vx] = $byte)
			$monitorValues[9] = "5XY0 Skip if Vx=Vy"
		Case 0x6000    ;6XNN Set $Vx to NN
			$monitorValues[8] = "Vx= " & $V[$Vx]
			$V[$Vx] = $byte
			$Pc = $Pc + 2
			$monitorValues[8] = $monitorValues[8] & " byte=" & $byte & " Result:" & $V[$Vx]
			$monitorValues[9] = "6XNN  Vx = byte"
		Case 0x7000    ;7XNN Adds NN to $Vx
			$V[$Vx] = BitAND(($V[$Vx] + $byte), 0xff)
			$Pc = $Pc + 2
			$monitorValues[8] = "Vx= " & $V[$Vx] & " byte=" & $byte & " Result:" & $V[$Vx]
			$monitorValues[9] = "7XNN Vx + Byte"
		Case 0x8000        ;Contains more data in last nibble

			Switch $value
				Case 0x00        ;8XY0: Sets $Vx to the value of VY
					$monitorValues[8] = "Vx= " & $V[$Vx]
					$V[$Vx] = $V[$Vy]
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx]
					$monitorValues[9] = "8XY0 Vx=Vy"

				Case 0x01        ;8XY1 Sets $Vx to $Vx or VY

					$monitorValues[8] = "Vx= " & $V[$Vx]
					$V[$Vx] = BitAND(BitOR($V[$Vx], $V[$Vy]), 0xff)
					If $Chip8Version = 1 Then $V[0xF] = 0 ;Fixing a Chip8 quirk
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx] & " Carry=" & $V[0xF]
					$monitorValues[9] = "8XY1 Vx=Vx OR Vy"

				Case 0x02        ;8XY2 Sets $Vx to $Vx and VY
					$monitorValues[8] = "Vx= " & $V[$Vx]
					$V[$Vx] = BitAND($V[$Vx], $V[$Vy])
					If $Chip8Version = 1 Then $V[0xF] = 0 ; Fixing a Chip8 quirk
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx] & " Carry=" & $V[0xF]
					$monitorValues[9] = "8XY2 Vx=Vx AND Vy"

				Case 0x03        ;8XY3 Sets $Vx To $Vx Xor Vy
					$monitorValues[8] = "Vx= " & $V[$Vx]
					$V[$Vx] = BitAND(BitXOR($V[$Vx], $V[$Vy]), 0xff)
					If $Chip8Version = 1 Then $V[0xF] = 0 ; Fixing a Chip8 quirk
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx] & " Carry=" & $V[0xF]
					$monitorValues[9] = "8XY3 Vx=Vx XOR Vy"
				Case 0x04        ;Adds VY to $Vx. VF is set to 1 when carry applies else to 0
					$monitorValues[8] = "Vx= " & $V[$Vx]
					$carry = (($V[$Vx] + $V[$Vy]) > 0xFF)
					$V[$Vx] = BitAND(($V[$Vx] + $V[$Vy]), 0xff)
					$V[0xF] = $carry
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx] & " Carry=" & $V[0xF]
					$monitorValues[9] = "8XY4 Vx=Vx + Vy"
				Case 0x05        ;VY is subtracted from $Vx. VF is set to 0 when there is a borrow else 1
					$carry = EqHi($V[$Vx], $V[$Vy])
					;$carry = ($V[$Vy] <=  $V[$Vx])
					$monitorValues[8] = "Vx= " & $V[$Vx]
					$V[$Vx] = BitAND(($V[$Vx] - $V[$Vy]), 0xff)
					$V[0xF] = $carry
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx] & " Carry=" & $V[0xF]
					$monitorValues[9] = "8XY5 Vx=Vx - Vy"

				Case 0x06        ;8XY6: Shift $Vx right by one, VF is set to the least significant bit of $Vx
					$monitorValues[8] = "Vx= " & $V[$Vx]
					If $Chip8Version = 1 Then
						$carry = BitAND($V[$Vy], 1)
						$V[$Vx] = BitShift($V[$Vy], 1)
					Else
						$carry = BitAND($V[$Vx], 1)
						$V[$Vx] = BitShift($V[$Vx], 1)
					EndIf
					$V[0xF] = $carry
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx] & " Carry=" & $V[0xF]
					$monitorValues[9] = "8XY6 Vx= Vx SHR 1  (Chip8 Vx= Vy SHR 1)"

				Case 0x07        ;8XY7 Sets $Vx to VY minus $Vx. VF is set to 0 when there's a borrow, and 1 when there isn't.
					;$carry = ($V[$Vx] <= $V[$Vy])
					$monitorValues[8] = "Vx= " & $V[$Vx]
					$carry = EqHi($V[$Vy], $V[$Vx])
					$V[$Vx] = BitAND(($V[$Vy] - $V[$Vx]), 0xff)
					$V[0xF] = $carry
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx] & " Carry=" & $V[0xF]
					$monitorValues[9] = "8XY7 Vx= Vy - $Vx"

				Case 0x0e        ;8XYE Shifts $Vx left by one. VF is set to the value of the most significant bit of $Vx before the shift.

					$monitorValues[8] = "Vx= " & $V[$Vx]
					If $Chip8Version = 1 Then
						$carry = (BitAND($V[$Vy], 0x80) > 0)
						$V[$Vx] = BitAND(BitShift($V[$Vy], -1), 0xff)
					Else
						$carry = (BitAND($V[$Vx], 0x80) > 0)
						$V[$Vx] = BitAND(BitShift($V[$Vx], -1), 0xff)
					EndIf
					$V[0xF] = $carry
					$Pc = $Pc + 2
					$monitorValues[8] = $monitorValues[8] & " Vy=" & $V[$Vy] & " Result:" & $V[$Vx] & " Carry=" & $V[0xF]
					$monitorValues[9] = "8XY7 Vx= Vx SHL 1 (Chip8 Vx= Vy SHL 1)"
				Case Else
					$monitorValues[7] = "8NNN - Undefined"
					$go = 0
					If $bPlay = True Then SNDPlay(0)
			EndSwitch

		Case 0x9000        ;9XY0 Skips the next $instruction if $Vx doesn't equal VY.
			$Pc = $Pc + 2 + (($V[$Vx] <> $V[$Vy]) * 2)
			$monitorValues[8] = "Vx= " & $V[$Vx] & " byte=" & $byte & " Result:" & ($V[$Vx] <> $byte)
			$monitorValues[9] = "9XY0 Skip if Vx<>Vy"
		Case 0xA000        ;ANNN set I to NNN
			$monitorValues[8] = "I=" & $I
			$I = $address
			$Pc = $Pc + 2
			$monitorValues[8] = $monitorValues[8] & " Result I=" & $I
			$monitorValues[9] = "ANNN Set I to Address"
		Case 0xB000        ;BNNN Jumps to the $address NNN plus V0.
			$monitorValues[8] = "V0 = " & $V[0]
			$Pc = $address + $V[0]
			$monitorValues[8] = $monitorValues[8] & " Address= " & $address & " Result=" & $Pc
			$monitorValues[9] = "BNNN Jump to Address + V0"
		Case 0xC000    ;CXNN Set $Vx to a random number and NN
			$tmp = Random(0, 255, 1)
			$monitorValues[8] = "Vx =" & $V[$Vx] & " byte=" & $byte & " RND=" & $tmp
			$V[$Vx] = BitAND($tmp, $byte)
			$Pc = $Pc + 2
			$monitorValues[8] = $monitorValues[8] & " Result:" & $V[$Vx]
			$monitorValues[9] = "CXNN Vx= (RND and Byte)"
		Case 0xD000    ;DXYN Draws a Sprite (x,y) size (8, N) Sprite is located at I
			$sprite = ""
			Global $sprite[17]

			$height = $value
			$V[0xF] = 0  ; Set the collision flag to 0

			$sprite[0] = $height

			For $y = 0 To $height - 1

				$ln = $mem[$I + $y]
				$sprite[$y + 1] = ""
				For $x = 0 To 7

					$pixel = BitAND($ln, BitShift(0x80, $x))

					$tmp = "."
					If $pixel <> 0 Then $tmp = Chr(0xdb)

					$sprite[$y + 1] = $sprite[$y + 1] & $tmp

					If $pixel <> 0 Then
						$tx = Mod($V[$Vx], 64)
						$ty = Mod($V[$Vy], 32)

						$tx = $tx + $x
						$ty = $ty + $y

						If $tx < 64 And $ty < 32 Then
							$id = ($ty * 64) + $tx
							If $Display[$id] = 1 Then $V[0xF] = 1 ; Collision happened
							$Display[$id] = BitXOR($Display[$id], 1)
						EndIf
					EndIf
				Next
			Next
			$needRedraw = True
			$Pc = $Pc + 2
			$monitorValues[8] = "Collision= " & $V[0xF]
			$monitorValues[9] = "DXYN Draw the Sprites"
		Case 0xE000
			Switch $byte
				Case 0x9E    ;EX9E	Skip the next $instruction if the $Vx is pressed
					$Pc = $Pc + 2 + (($keys[$V[$Vx]] > 0) * 2)
					$monitorValues[8] = "Vx=" & $V[$Vx]
					$monitorValues[9] = "EX9E Skip if key VX is pressed"
				Case 0xA1    ;;EXA1	Skip the next $instruction if the $Vx is NOT pressed
					$Pc = $Pc + 2 + ((($keys[$V[$Vx]]) = 0) * 2)
					$monitorValues[8] = "Vx=" & $V[$Vx]
					$monitorValues[9] = "EXA1 Skip if key VX is NOT pressed"
				Case Else
					$monitorValues[7] = "ENNN - Undefined"
					$go = 0
					If $bPlay = True Then SNDPlay(0)
			EndSwitch

		Case 0xF000            ;Multicase statements

			Switch $byte
				Case 0x07        ;FX07: Set VX to the value of delay_timer
					$V[$Vx] = $DelayTimer
					$Pc = $Pc + 2
					$monitorValues[8] = "DelayTimer= " & $DelayTimer
					$monitorValues[9] = "FX07 Set Vx to Delay Timer"
				Case 0x0a        ;FX0A A key press is awaited, and then stored in VX.

					If $anykeypressed = -2 Then
						$monitorValues[8] = "Status: Waiting for a key"
						For $x = 0 To 15
							If $keys[$x] = 1 Then
								$anykeypressed = $x
								$monitorValues[8] = "Status: Key pressed, Nr=" & $anykeypressed
								ExitLoop
							EndIf
						Next
					Else
						$monitorValues[8] = "Status: Waiting for the Release of " & $anykeypressed
						If $keys[$anykeypressed] = 0 Then
							$V[$Vx] = $anykeypressed
							$anykeypressed = -2
							$Pc = $Pc + 2    ;A key has been released, skip the instruction
							$monitorValues[8] = "Status: Key released, Vx=" & $anykeypressed
						EndIf
					EndIf

					$monitorValues[9] = "FX0a Wait for A Keypress"
				Case 0x15        ;FX15: Set delay timer to V[x]
					$DelayTimer = $V[$Vx]
					$Dtimer = TimerInit()
					$Pc = $Pc + 2
					$monitorValues[9] = "FX15 DelayTimer=Vx"
				Case 0x18        ;FX18: Set the sound timer to V[x]
					$SoundTimer = $V[$Vx]
					$Stimer = TimerInit()
					$Pc = $Pc + 2
					$monitorValues[9] = "FX18 SoundTimer=Vx"
				Case 0x1E        ;FX1E: Adds VX to I
					$carry = ($I + $V[$Vx]) > 0xFFF
					$I = BitAND(($I + $V[$Vx]), 0xFFF)
					$V[0xf] = $carry
					$Pc = $Pc + 2
					$monitorValues[9] = "FX1E SoundTimer=Vx"

				Case 0x29    ;Sets I to the location of the Sprite for the character $Vx (fontset)

					$I = $FontStartAddress + ($V[$Vx] * 5)
					$Pc = $Pc + 2
					$monitorValues[8] = "Vx=" & $V[$Vx] & + " FontStart=" & $FontStartAddress & " *5 =" & ($V[$Vx] * 5)
					$monitorValues[9] = "FX29 I=Font nr"
				Case 0x33    ;FX33 store a binary coded decimal value $Vx in I, I+1 and I+2

					$B100 = Int($V[$Vx] / 100)
					$B10 = Int(Mod(($V[$Vx] / 10), 10))
					$B1 = Int(Mod(Mod($V[$Vx], 100), 10))
					$mem[$I] = $B100
					$mem[$I + 1] = $B10
					$mem[$I + 2] = $B1
					$Pc = $Pc + 2
					$monitorValues[8] = "Vx=" & $V[$Vx] & " I=" & $B100 & " I+1= " & $B10 & " I+2= " & $B1
					$monitorValues[9] = "FX33 BCD of Vx to I,I+1,I+2"

				Case 0x55        ;FX55 Stores V0 to VX in memory starting at address I.
					For $t = 0 To $Vx
						$mem[$I + $t] = $V[$t]
						$monitorValues[8] = $monitorValues[8] & $V[$t] & ";"
					Next
					if $ChipQuirkFix=0 then $I = $I + $Vx + 1
					$Pc = $Pc + 2

					$monitorValues[9] = "FX55 Store V0...Vx Into Memory"

				Case 0x65    ;FX65 Fills V0 To $Vx with values from I
					For $t = 0 To $Vx
						$V[$t] = $mem[$I + $t]
					Next
					if $ChipQuirkFix=0 then  $I = $I + $Vx + 1
					$Pc = $Pc + 2
					$monitorValues[9] = "FX65 V0...Vx from Mem I+x"

				Case Else
					$monitorValues[7] = "FNNN - Undefined"
					$go = 0
					If $bPlay = True Then SNDPlay(0)
			EndSwitch

		Case Else
			$monitorValues[7] = "Global Instruction Undefined"
			$go = 0
			If $bPlay = True Then SNDPlay(0)

	EndSwitch

	If $SoundTimer > 0 Then

		If TimerDiff($Stimer) > (40 + ($speed * 2)) Then
			$SoundTimer = $SoundTimer - 1
			$Stimer = TimerInit()
			If $bPlay = False Then SNDPlay(1)
		EndIf
	Else
		If $bPlay = True Then SNDPlay(0)
	EndIf

	If $DelayTimer > 0 Then
		If TimerDiff($Dtimer) > (40 + ($speed * 2)) Then
			$DelayTimer = $DelayTimer - 1
			$Dtimer = TimerInit()
		EndIf
	EndIf

	If $Pc = $loopdetect_oldpc Then
		If $instruction = 0x1000 Or $instruction = 0xB000 Then
			$loopdetect_count = $loopdetect_count + 1
			If $loopdetect_count > 2 Then
				$go = 0
				If $bPlay = True Then SNDPlay(0)
				$loopdetect_count = 0
				$monitorValues[7] = "Stopped:Infinite Loop " & Hex($Pc, 3)
			EndIf
		EndIf
	Else
		$loopdetect_count = 0
	EndIf

	If $DBGWinOn = 1 Then
		$tmp1 = 0
		$tmp2 = -1
		$tmp3 = 0
		If _IsChecked($dbgeditCheckActive) Then
			If _IsChecked($dbgeditRPC) Then $tmp1 = 1 ;PC radio button
			If _IsChecked($dbgeditRCY) Then $tmp1 = 2 ;Cycle radio
			If _IsChecked($dbgeditRVR) Then $tmp1 = 3 ;Vreg radio
			If _IsChecked($dbgeditRIn) Then $tmp1 = 4 ;Instruction radio
			If _IsChecked($dbgeditRI) Then $tmp1 = 5 ;I radio
			If _IsChecked($dbgeditRST) Then $tmp1 = 6 ;Stack radio
			Switch $tmp1
				Case 1                ;Stop when PC = Input Nr
					$tmp2 = GUICtrlRead($dbgeditInput1)
					If $Pc = $tmp2 Then
						$go = 0
						$monitorValues[7] = "Breakpoint at PC"
					EndIf

				Case 2                ;Stop at cycle NR
					$tmp2 = GUICtrlRead($dbgeditInput1)
					If $cycle = $tmp2 Then
						$go = 0
						$monitorValues[7] = "Breakpoint at Cycle"
					EndIf

				Case 3                ;Stop if V(x) = Input
					$tmp2 = GUICtrlRead($dbgeditInput1)
					$tmp3 = GUICtrlRead($dbgeditCoVr)
					If $V[$tmp3] = $tmp2 Then
						$go = 0
						$monitorValues[7] = "Breakpoint at V ( " & $tmp3 & " )"

					EndIf
				Case 4
					$tmp = "12345679ABCD"
					$tmp1 = StringLeft(Hex($instruction, 4), 1)
					$tmp2 = StringLeft(GUICtrlRead($dbgeditCoIn), 1)
					If StringInStr($tmp, $tmp1) Then
						If $tmp1 = $tmp2 Then
							$go = 0
							$monitorValues[7] = "Breakpoint at Instruction " & $tmp1
						EndIf
					ElseIf StringInStr("0EF", $tmp1) Then
						$tmp3 = StringRight(GUICtrlRead($dbgeditCoIn), 2)
						If Hex($byte, 2) = $tmp3 Then
							$go = 0
							$monitorValues[7] = "Breakpoint at Instruction " & $tmp3
						EndIf
					ElseIf $tmp1 = "8" Then
						$tmp3 = StringRight(GUICtrlRead($dbgeditCoIn), 1)
						If Hex($value, 1) = $tmp3 Then
							$go = 0
							$monitorValues[7] = "Breakpoint at Instruction " & $tmp3
						EndIf
					EndIf
				Case 5
					If $I = GUICtrlRead($dbgeditInput1) Then
						$go = 0
						$monitorValues[7] = "Breakpoint at I"
					EndIf
				Case 6
					If $StackPointer > 0 And $StackPointer <= 16 Then
						If $Stack[$StackPointer - 1] = GUICtrlRead($dbgeditInput1) Then
							$go = 0
							$monitorValues[7] = "Breakpoint at Stack = " & $Stack[$StackPointer - 1]
						EndIf
					EndIf
			EndSwitch
			$tmp = ""
			$tmp1 = ""
			$tmp2 = ""
			$tmp3 = ""
			If $go = 0 And $bPlay = True Then SNDPlay(0)
		EndIf

	EndIf

	;If $needRedraw = True Then DrawScreen()  ; For GuiOnEvent Mode
EndFunc   ;==>ChipRun

Func _IsChecked($idControlID)
	;Return BitAND(GUICtrlRead($idControlID), $GUI_CHECKED) = $GUI_CHECKED     ;Returns true or false (oneliner)
	;The lines below convert true and false to numbers - 1 and 0
	Local $x = BitAND(GUICtrlRead($idControlID), $GUI_CHECKED) = $GUI_CHECKED
	If $x = True Then Return 1
	Return 0
EndFunc   ;==>_IsChecked

Func BitRead($x, $bit)               ;Reads a bit state of a variable
	;Return the bit state (0 or 1) rightmost bit being 1st
	Return ((BitAND(($x), (BitShift(1, -($bit - 1))))) <> 0) ? 1 : 0
EndFunc   ;==>BitRead

Func ChangePixelType($nr = 0)
	Local $tmp = Chr(219)
	If $nr = 0 Then
		$PixNr = Mod($PixNr + 1, 7)
	Else
		$PixNr = $nr
	EndIf

	If $PixNr = 0 Then $tmp = Chr(147)
	If $PixNr = 1 Then $tmp = Chr(219)
	If $PixNr = 2 Then $tmp = "*"
	If $PixNr = 3 Then $tmp = "@"
	If $PixNr = 4 Then $tmp = "O"
	If $PixNr = 5 Then $tmp = "o"
	If $PixNr = 6 Then $tmp = "#"
	Return $tmp
EndFunc   ;==>ChangePixelType

Func DrawScreen()
	Local $tmp = ""  ;cls

	Local $c = 0, $x, $y

	For $y = 0 To 31
		For $x = 0 To 63
			If $Display[$c] = 1 Then
				$tmp = $tmp & $Pix
			Else
				$tmp = $tmp & " "
			EndIf
			$c = $c + 1
		Next
		$tmp = $tmp & @CRLF
	Next

	;GUICtrlSetData($DisplayLabel, $tmp)
	ControlSetText("", "", $DisplayLabel, $tmp)
	;GUICtrlSetData($idEdit, $tmp)
	$needRedraw = False
EndFunc   ;==>DrawScreen

Func ClearDisplay()
	For $x = 0 To 64 * 32
		$Display[$x] = 0
	Next
EndFunc   ;==>ClearDisplay

Func ClearKeys()
	For $x = 0 To 15
		$keys[$x] = 0
	Next
EndFunc   ;==>ClearKeys

Func GetTheKeys()
	ClearKeys()
	If WinActive($Gui_hWnd) Then
		For $x = 0 To 15
			If _IsPressed($keyNr[$x]) = 1 Then $keys[$x] = 1
		Next
	EndIf

	If _IsPressed("26") Then $keys[$KeySets[$KSN][0]] = 1
	If _IsPressed("25") Then $keys[$KeySets[$KSN][1]] = 1
	If _IsPressed("27") Then $keys[$KeySets[$KSN][2]] = 1
	If _IsPressed("28") Then $keys[$KeySets[$KSN][3]] = 1
	If _IsPressed("11") Then $keys[$KeySets[$KSN][4]] = 1
	If _IsPressed("12") Then $keys[$KeySets[$KSN][5]] = 1
	If _IsPressed("10") Then $keys[$KeySets[$KSN][6]] = 1

	;If _IsPressed(59) Then $Keys[10]=1 ;add extra z key (in case they are swapped)
EndFunc   ;==>GetTheKeys

Func Init($clrmem = 0)
	$I = 0
	$Pc = 0x200
	$StackPointer = 0
	$DelayTimer = 0
	$SoundTimer = 0
	$cycle = 0
	$anykeypressed = -2

	For $x = 0 To 15
		$V[$x] = 0
	Next
	For $x = 0 To 15
		$Stack[$x] = 0
	Next

	If $clrmem = 1 Then
		For $x = 0x200 To $totalmemory - 1
			$mem[$x] = 0
		Next
	Else
		For $x = 0 To $totalmemory - 1
			$mem[$x] = $MemCopy[$x]
		Next
	EndIf

	ClearKeys()
	ClearDisplay()
	DrawScreen()
EndFunc   ;==>Init

Func UpdateInfo()
	If $go = 0 And $bPlay = True Then SNDPlay(0)
	WinSetTitle($Gui_hWnd, "", "Chip 8 - " & $RomName)
	ControlSetText($Gui_hWnd, "", $input, $edittext)
	Sleep(20)
EndFunc   ;==>UpdateInfo

Func ExtractFilename($fn)
	Local $p = 0
	Local $ln = StringLen($fn) - 1
	Local $c = ""

	For $x = $ln To 1 Step -1
		$c = StringMid($fn, $x, 1)
		If $c = "\" Or $c = ":" Then ExitLoop
	Next
	$c = StringMid($fn, $x + 1)
	Return $c
EndFunc   ;==>ExtractFilename

Func LoadChip($fname, $newPc = 0x200)
	If $newPc > $totalmemory - 1 Then
		$monitorValues[7] = "Load Error: above mem limit"
		Return -1 ; Error, start of the loading address is above the memory limit
	EndIf
	Local $S, $tmp, $sl
	If FileExists($fname) = 1 Then
		$S = FileGetSize($fname)
		;ConsoleWrite("size: " & $S & " Total-NewPC= " & ($totalmemory - $newPc) & @CRLF)
		If $S > 0 And $S < ($totalmemory - $newPc) Then
			$FH = FileOpen($fname, $FO_READ + $FO_BINARY)
			If $FH < 0 Then
				$monitorValues[7] = "Failed to open file"
				Return -2
			EndIf
			$F = FileRead($FH)
			FileClose($FH)
			$F = StringMid($F, 3)
			;ConsoleWrite($F & @CRLF)
			$sl = StringLen($F) / 2
			For $x = 0 To $sl
				$tmp = StringMid($F, ($x * 2) + 1, 2)
				If $newPc + $x < $totalmemory Then
					$mem[$newPc + $x] = Dec($tmp)
				Else
					;ConsoleWrite ("New PC=" & $newPc & " x=" & $x & " StringLen=" & $sl & "Filesize=" & $S & @CRLF)
					$monitorValues[7] = "Memory overflow"
				EndIf
			Next
			$RomName = ExtractFilename($fname)
			For $x = 0 To $totalmemory - 1
				$MemCopy[$x] = $mem[$x]
			Next
		Else
			$monitorValues[7] = "Invalid filesize"
			Return -3            ;Error, invalid filesize
		EndIf
	Else
		$monitorValues[7] = "File does not exists"
		Return -4                ;Error, file does not exist
	EndIf

	Return 0
EndFunc   ;==>LoadChip

Func EqHi($a, $b)
	If $a >= $b Then Return 1
	Return 0
EndFunc   ;==>EqHi

Func AddDefaultFont()
	$PokeStart = $FontStartAddress
	PokeHexToRom("F0909090F02060202070F010F080F0F010F010F09090F01010F080F010F0F080F090F0F010204040")
	PokeHexToRom("F090F090F0F090F010F0F090F09090E090E090E0F0808080F0E0909090E0F080F080F0F080F08080")
EndFunc   ;==>AddDefaultFont


Func AddTestRom()
	$RomName = "Built-It Rom: "
	$PokeStart = 0x200
	Local $r = Random(0, 3, 1)
	If $r = 0 Then
		$RomName = $RomName & "Smiling Face"
		PokeHexToRom("601E610A630162016705680969006A076B01A21A2261224F12146600817E3C410000EE81B54100620100EE411A00EE7101411A620000EE400000EE80B54000630100EE403800EE70014038630000EEE7A1")
		PokeHexToRom("121FE8A11243E9A1122BEAA1123700EE00E0D01500EE")
	ElseIf $r = 1 Then
		$RomName = $RomName & "Chip 8 Picture"
		PokeHexToRom("00E0A2486000611E6200D202D21272083240120A6000613E6202A24AD02ED12E720ED02ED12EA258600B6108D01F700AA267D01F700AA276D01F7003A285D01F700AA294D01F1246FFFFC0C0C0C0C0C0C0")
		PokeHexToRom("C0C0C0C0C0C0C0FF80808080808080808080808080FF81818181818181FF81818181818181808080808080808080808080808080FF818181818181FF80808080808080FF818181818181FF818181818181FFFF")
	ElseIf $r = 2 Then
		$RomName = $RomName & "Brix [Andreas Gustafsson]"
		PokeHexToRom("6E0565006B066A00A30CDAB17A043A4012087B023B1212066C206D1FA310DCD122F660006100A312D0117008A30ED0116040F015F00730001234C60F671E680169FFA30ED671A310DCD16004E0A17CFE60")
		PokeHexToRom("06E0A17C02603F8C02DCD1A30ED67186848794603F8602611F8712471F12AC46006801463F68FF47006901D6713F0112AA471F12AA600580753F0012AA6001F018806061FC8012A30CD07160FE890322F6")
		PokeHexToRom("750122F6456012DE124669FF806080C53F0112CA610280153F0112E080153F0112EE80153F0112E86020F018A30E7EFF80E080046100D0113E00123012DE78FF48FE68FF12EE7801480268016004F01869")
		PokeHexToRom("FF1270A314F533F265F12963376400D3457305F229D34500EEE0008000FC00AA0000000000")
	Else
		$RomName = $RomName & "Maze (David Winter)"
		PokeHexToRom("60006100A222C2013201A21ED0147004304012046000710431201204121C8040201020408010")
	EndIf

	For $x = 0 To $totalmemory - 1
		$MemCopy[$x] = $mem[$x]
	Next

EndFunc   ;==>AddTestRom

Func PokeHexToRom($hexstring)
	;Does nothing if the hexstring is not even

	Local $L = StringLen($hexstring), $y
	If $L = 0 Then
		$monitorValues[7] = "PokeHex StringLen = 0"
		Return -1         ; Error 0 length
	EndIf
	If IsEven(StringLen($hexstring)) = 0 Then
		$monitorValues[7] = "PokeHex - Uneven bytes"
		Return -2 ; Error Not even (needs pairs of bytes)
	EndIf
	For $x = 0 To $L - 1 Step 2
		$y = Dec(StringMid($hexstring, $x + 1, 2))
		$mem[$PokeStart] = $y
		$PokeStart = $PokeStart + 1
		If $PokeStart > 4096 And ($x < $L) Then
			$monitorValues[7] = "PokeHex - Start add too high"
			Return -3 ;error Starting address is now above
		EndIf
	Next
EndFunc   ;==>PokeHexToRom

Func IsEven($nr)
	Return 1 - BitAND($nr, 1)
EndFunc   ;==>IsEven

Func SetInfoFont($nr)
	If $nr = 0 Then
		GUICtrlSetFont($Info_Label, 11, $FW_Normal, $GUI_FONTNORMAL, "Terminal", $DRAFT_QUALITY)
	Else
		GUICtrlSetFont($Info_Label, 8, $FW_Normal, $GUI_FONTNORMAL, "C64 Pro Mono") ;Testing
	EndIf
EndFunc   ;==>SetInfoFont

Func LoadSound($sWav = @ScriptDir & '\beep.wav')
	; Read a .wav to memory
	If $sounddriver = 0 Then
		Local $dWav = FileRead($sWav)
		If @error Then
			$monitorValues[7] = "Error loading Sound file"
			Return 0 ;'Error Unable to read "' & $sWav & '"'
		EndIf
		$tWav = DllStructCreate('byte[' & BinaryLen($dWav) & ']')
		DllStructSetData($tWav, 1, $dWav)
		$pWav = DllStructGetPtr($tWav)
	Else
		$pWav = _SoundOpen($sWav)
	EndIf
EndFunc   ;==>LoadSound

;_SoundOpen

Func SNDPlay($pl = 1)

	If $sounddriver = 0 Then
		$bPlay = Not $bPlay
		If $pl = 0 Then $bPlay = False
		If $bPlay = True Then
			If $MuteSound = 0 Then _WinAPI_PlaySound($pWav, BitOR($SND_ASYNC, $SND_LOOP, $SND_MEMORY))
		Else
			_WinAPI_PlaySound('')
		EndIf
	Else
		$bPlay = Not $bPlay
		If $pl = 0 Then $bPlay = False
		If $bPlay = True Then
			If $MuteSound = 0 Then _SoundPlay($pWav, 0)
		Else
			_SoundStop($pWav)
		EndIf
	EndIf
EndFunc   ;==>SNDPlay

Func KeyConfig()

	Local $tmp, $tmp1

	#Region ### START Koda GUI section ### Form=
	$KeyConfigForm = GUICreate("Extra Button Config", 329, 140, -1126, 187)
	$KCC[0] = GUICtrlCreateCombo("Up", 76, 2, 78, 25, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
	GUICtrlSetTip(-1, "Arrow Up")
	$KCC[1] = GUICtrlCreateCombo("Left", 0, 33, 78, 28, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
	GUICtrlSetTip(-1, "Arrow Left")
	$KCC[2] = GUICtrlCreateCombo("Right", 154, 33, 78, 28, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
	GUICtrlSetTip(-1, "Arrow Right")
	$KCC[3] = GUICtrlCreateCombo("Down", 77, 64, 78, 28, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
	GUICtrlSetTip(-1, "Arrow Down")
	$KCC[4] = GUICtrlCreateCombo("B1", 243, 2, 78, 28, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
	GUICtrlSetTip(-1, "Ctrl")
	$KCC[5] = GUICtrlCreateCombo("B2", 243, 33, 78, 28, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
	GUICtrlSetTip(-1, "Alt")
	$KCC[6] = GUICtrlCreateCombo("B3", 243, 64, 78, 28, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
	GUICtrlSetTip(-1, "Shift")
	Global $ComboCPreset = GUICtrlCreateCombo("", 1, 115, 157, 25, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
	GUICtrlSetTip(-1, "Meaning of the numbers/letters:" & @CRLF & "* - Cant be saved (to the ini file) nor changed " & @CRLF & "Numbers mean, from left to right:" & @CRLF & _
			"Arrow (Up, Left, Right, Down) Ctrl Alt Shift", "Preset info")
	Global $ButtonCUse = GUICtrlCreateButton("Use", 189, 112, 62, 22)
	GUICtrlSetTip(-1, "Use the current config as the slot number, without saving it" & @CRLF & " (clicking on this will close the key config window)")
	Global $ButtonCSave = GUICtrlCreateButton("Save", 260, 112, 62, 22)
	GUICtrlSetTip(-1, "Use and Save the current slot number to the ini file" & @CRLF & "Attention: it will overwrite the current slot without warnings")
	GUICtrlCreateLabel("Please select a preset before changing the keys !", 0, 90, 250, 16)
	#EndRegion ### END Koda GUI section ###

	For $x = 0 To 6
		GUICtrlSetData($KCC[$x], "")
		GUICtrlSetData($KCC[$x], "x|1|2|3|q|w|e|a|s|d|y|c|4|r|f|v", "x")
	Next

	UpdateComboBox($ComboCPreset, $KCC)

	$tmp1 = ""
	$tmp = _GUICtrlComboBox_GetCurSel($ComboCPreset)

	If $tmp < 2 Then
		GUICtrlSetState($ButtonCSave, $GUI_DISABLE)
		_ComboboxDisableEnableAll($KCC, 0)
	EndIf
EndFunc   ;==>KeyConfig

Func UpdateComboBox(ByRef $combodest, ByRef $array)
	Local $tmp, $tmp1
	GUICtrlSetData($combodest, "")

	For $x = 0 To 9
		$tmp = "  "
		If $x < 2 Then
			$tmp = "* "
			If $x = 0 Then $tmp1 = "3weaqqq"
			If $x = 1 Then $tmp1 = "2qeswww"
		Else
			$tmp1 = IniRead($iniFile, "keyconfig", $x, "2qeswww")
		EndIf
		GUICtrlSetData($combodest, $tmp & $x & ": " & $tmp1)
		if $x=$KSN  then MakeKeyConfigText($tmp1)
		For $y = 0 To 6
			$KeySets[$x][$y] = StringInStr($KeyN, StringMid($tmp1, ($y + 1), 1)) - 1
			If $x = $KSN Then _GUICtrlComboBox_SetCurSel($array[$y], $KeySets[$KSN][$y])
		Next
	Next
	_GUICtrlComboBox_SetCurSel($combodest, $KSN)
EndFunc   ;==>UpdateComboBox

Func ReadCustomKeys()
	For $x = 0 To 9
		If $x < 2 Then
			If $x = 0 Then $tmp1 = "3weaqqq"
			If $x = 1 Then $tmp1 = "2qeswww"
		Else
			$tmp1 = IniRead($iniFile, "keyconfig", $x, "2qes134")
		EndIf
		$KeyConfigInfoTxt=""
		if $x=$KSN  then MakeKeyConfigText($tmp1)
		For $y = 0 To 6
			$KeySets[$x][$y] = StringInStr($KeyN, StringMid($tmp1, ($y + 1), 1)) - 1
		Next
	Next
EndFunc   ;==>ReadCustomKeys

Func MakeKeyConfigText($string)
	Local $tmp
	$tmp=@CRLF & "Custom Key Configuration:" & @CRLF & @CRLF
	$tmp=$tmp & "      Up:" & StringMid($string, 1, 1) & "           Ctrl :" & StringMid($string, 5, 1) & @CRLF
	$tmp=$tmp & "Left:" & StringMid($string, 2, 1) & "      Right:" & StringMid($string, 3, 1) & "  Alt  :" & StringMid($string, 6, 1)& @CRLF
	$tmp=$tmp & "      Down:" & StringMid($string, 4, 1) & "         Shift:" & StringMid($string, 7, 1) & @CRLF
    $KeyConfigInfoTxt=$tmp
EndFunc

Func _ComboboxDisableEnableAll(ByRef $array, $do = 0)
	Local $tmp
	If $do = 0 Then
		$tmp = $GUI_DISABLE
	Else
		$tmp = $GUI_ENABLE
	EndIf
	For $x = 0 To 6
		GUICtrlSetState($array[$x], $tmp)
	Next
EndFunc   ;==>_ComboboxDisableEnableAll


Func MY_SYS_COMMAND($hWnd, $msg, $wParam, $lParam)
	; functions used from https://www.autoitscript.com/forum/topic/134878-guiregistermsg-replacement-for-guictrlsetonevent-and-guigetmsg/page/2/#comments
	Local $nNotifyCode = BitShift($wParam, 16)
	Local $nID = BitAND($wParam, 0x0000FFFF)
	Local $hCtrl = $lParam
	Select
		Case $hWnd = $Gui_hWnd
			Switch $wParam
				Case 0xf060
					DllClose($hDLL)
					If $sounddriver = 0 Then
						_WinAPI_PlaySound('')
						$dWav = ""
						$tWav = ""
						$pWav = ""
					Else
						_SoundStop($pWav)
						_SoundClose($pWav)
					EndIf
					Exit
				Case 0xF120 ;ConsoleWrite("!Restore window" & @LF)
				Case 0xF020 ;ConsoleWrite("!Minimize Window" & @LF)
			EndSwitch
		Case $hWnd = $KeyConfigForm
			Switch $wParam
				Case 0xf060
					GUISetState(@SW_HIDE, $KeyConfigForm)
			EndSwitch
	EndSelect
	Return $GUI_RUNDEFMSG ; Only workout clicking on the button
EndFunc   ;==>MY_SYS_COMMAND



Func MY_WM_COMMAND($hWnd, $msg, $wParam, $lParam)
	Local $nNotifyCode = BitShift($wParam, 16)
	Local $nID = BitAND($wParam, 0x0000FFFF)
	Local $hCtrl = $lParam
	Switch $nID
		Case $BU_PP
			$infoChanged = 1
			$InfoNr = $InfoNr - 1
			Sleep(250)
		Case $BU_NP
			$infoChanged = 1
			$InfoNr = $InfoNr + 1
			Sleep(250)
		Case $BU_F
			If $InfoNr = 4 Then
				$MemTopAddr = $FontStartAddress
			Else
				SetInfoFont(1)
			EndIf
		Case $BU_P
			If $InfoNr = 4 Then
				$MemTopAddr = 0x200
			Else
				SetInfoFont(0)
			EndIf
		Case $BU_PC
			If $InfoNr = 4 Then
				$MemTopAddr = $Pc
				If $MemTopAddr < 0 Then $MemTopAddr = 0
				If $MemTopAddr > ($memstop) Then $MemTopAddr = $memstop
			Else
				$Pix = ChangePixelType(0)
				$needRedraw = 1
			EndIf
		Case $BU_B
			If $InfoNr = 4 Then
				$MemTopAddr = $memstop
			Else
				Local $aWP = WinGetPos($Gui_hWnd)
				WinMove($dbggui, "", $aWP[0] + 350, $aWP[1])
				$DBGOn = 1
				$aWP=""
			EndIf
		Case $BU_PD
			If $InfoNr = 4 Then
				$MemTopAddr = $MemTopAddr + ($MemDisplay * $memLines)
				If $MemTopAddr > ($memstop) Then $MemTopAddr = $memstop
				Sleep(180)
			Else
				Local $aWP = WinGetPos($Gui_hWnd)
				WinMove($KeyConfigForm, "", $aWP[0] + 380, $aWP[1]+80)
				GUISetState(@SW_SHOW, $KeyConfigForm)
				$aWP=""
				WinActivate ($KeyConfigForm)
			EndIf
		Case $BU_D
			If $InfoNr = 4 Then
				$MemTopAddr = $MemTopAddr + $MemDisplay
				If $MemTopAddr > ($memstop) Then $MemTopAddr = $memstop
			EndIf
		Case $BU_U
			If $InfoNr = 4 Then
				$MemTopAddr = $MemTopAddr - $MemDisplay
				If $MemTopAddr < 0 Then $MemTopAddr = 0
				Sleep(100)
			EndIf
		Case $BU_PU
			If $InfoNr = 4 Then
				$MemTopAddr = $MemTopAddr - ($MemDisplay * $memLines)
				If $MemTopAddr < 0 Then $MemTopAddr = 0
			EndIf
		Case $BU_T
			If $InfoNr = 4 Then
				$MemTopAddr = 0
			Else
				ShellExecute($iniFile)
			EndIf
		Case $Bu_Start, $dbgeditRun
			$go = Mod($go + 1, 2)
			UpdateInfo()
			Sleep(200)
		Case $Bu_Step
			$step = 1
			$go = 0
			UpdateInfo()
			Sleep(50)
		Case $Bu_SpeedM
			$speed = $speed - 1
			If $speed < 0 Then $speed = 0
			UpdateInfo()
		Case $Bu_SpeedP
			$speed = $speed + 1
			If $speed > 20 Then $speed = 20
			UpdateInfo()
		Case $Bu_BGC
			$tmp = _ChooseColor(2)
			$Colors[1] = $tmp
			GUICtrlSetBkColor($DisplayLabel, $Colors[1])
			IniWrite($iniFile, "config", "ColorBG", $Colors[1])
		Case $Bu_FGC
			$tmp = _ChooseColor(2)
			$Colors[0] = $tmp
			GUICtrlSetColor($DisplayLabel, $Colors[0])
			IniWrite($iniFile, "config", "ColorFG", $Colors[0])
		Case $Bu_Reset
			Init()
			$go = 0
			UpdateInfo()
		Case $Bu_Reboot
			Init(1)
			AddTestRom()
			$go = 1
			UpdateInfo()
		Case $Bu_Chip8
			$Chip8Version = Mod($Chip8Version + 1, 2)
			UpdateInfo()
		Case $Bu_QuirkFix
			$ChipQuirkFix=Mod($ChipQuirkFix+1,2)
			UpdateInfo()
		Case $Bu_Mute
			$MuteSound = Mod(($MuteSound + 1), 2)
			UpdateInfo()
			IniWrite($iniFile, "config", "MuteSound", $MuteSound)
		Case $Bu_Load
			SNDPlay(0)
			$fname = FileOpenDialog("Select a Chip8 file (.ch8)", $fname, "Chip 8 Files (*.CH8)|All Files (*.*)")
			Init(1)
			LoadChip($fname)
			UpdateInfo()
			ControlFocus($Gui_hWnd, "", $input)
		Case $dbgeditCLS
			GUICtrlSetData($dbgedit, "")
		Case $dbgeditHide
			$DBGOn = 0
		Case $dbgeditInputDec
			GUICtrlSetData($dbgeditInputHex, Hex(GUICtrlRead($dbgeditInputDec)))
		Case $dbgeditInputHex
			GUICtrlSetData($dbgeditInputDec, Dec(GUICtrlRead($dbgeditInputHex)))
		Case $dbgeditUSE
			GUICtrlSetData($dbgeditInput1, GUICtrlRead($dbgeditInputDec))
		Case $ComboCPreset
			$NrComboBoxPreset = _GUICtrlComboBox_GetCurSel($ComboCPreset)
			For $y = 0 To 6
				_GUICtrlComboBox_SetCurSel($KCC[$y], $KeySets[$NrComboBoxPreset][$y])
			Next
			If $NrComboBoxPreset < 2 Then
				GUICtrlSetState($ButtonCSave, $GUI_DISABLE)
				_ComboboxDisableEnableAll($KCC, 0)
			Else
				GUICtrlSetState($ButtonCSave, $GUI_ENABLE)
				_ComboboxDisableEnableAll($KCC, 1)
			EndIf
		Case $ButtonCSave
			$KSN = $NrComboBoxPreset
			$tmp1 = ""
			For $x = 0 To 6
				$KeySets[$KSN][$x] = _GUICtrlComboBox_GetCurSel(($KCC[$x]))
				$tmp1 = $tmp1 & StringMid($KeyN, $KeySets[$NrComboBoxPreset][$x] + 1, 1)
			Next
			IniWrite($iniFile, "keyconfig", $NrComboBoxPreset, $tmp1)
			$tmp1 = ""
			UpdateComboBox($ComboCPreset, $KCC)
		Case $ButtonCUse
			$KSN = $NrComboBoxPreset
			$tmp=""
			For $x = 0 To 6
				$KeySets[$KSN][$x] = _GUICtrlComboBox_GetCurSel(($KCC[$x]))
				$tmp=$tmp & StringMid($KeyN,$KeySets[$KSN][$x]+1, 1)
			Next
			MakeKeyConfigText($tmp)
			$tmp=""
			GUISetState(@SW_HIDE, $KeyConfigForm)
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc   ;==>MY_WM_COMMAND


; Use as an anlternative keyboard binding ?
;~ "     Nr  Assigned   Literal      Arrow Keys
;~ "Vx=   0    x          58
;~ "Vx=   1    1          31
;~ "Vx=   2    2          32          Up    26
;~ "Vx=   3    3          33
;~ "Vx=   4    q          51          Left  25
;~ "Vx=   5    w          57          Ctrl  11
;~ "Vx=   6    e          45          Right 27
;~ "Vx=   7    a          41
;~ "Vx=   8    s          53          Down  28
;~ "Vx=   9    d          44
;~ "Vx=   10   y          59
;~ "Vx=   11   c          43
;~ "Vx=   12   4          34
;~ "Vx=   13   r          52
;~ "Vx=   14   f          46
;~ "Vx=   15   v          56
;
;Global $keyNr[] = [58, 31, 32, 33, 51, 57, 45, 41, 53, 44, 59, 43, 34, 52, 46, 56]
;Key Meanings     " x    1  2   3   q   w   e   a    s   d   y   c   4    r  f   v"
;Global $keyNrExtra[]=

;"x123qweasdyc4rfv"
