
;V 1.00 mod for the chip8 emu
;Ascii Game Framework - include, written by Dan_555
;Use this to easily program some Ascii-Games
;
;Initialize(Height,Width,"windowname")	to startup the GDI+ and to setup default color and font.Defaults to 400x300 window
;CleanUp()  							To free the resources used by the script.
;
;The available commands are:
;
;CLS  ("000000")			Clears the screen in a color. The color is in RGB - Hex format.

;Color ("000000")			Sets the brush color in RGB - Hex format. Optionally, the alpha value "00"-"FF" can be set up, but it is predefined as "FF".

;Flip()						Draws the buffer to the window (flips the backbuffer so that it becomes visible) use this at the end of the game loop.

;Available Global variables:
;
;$gdip_GraphicX, $gdip_GraphicY					Current Drawing size
;$mouseX, $mouseY, $mouseGX, $mouseGY           Mouse Coordinates x,y=Window based (-1,-1 if outside window) GX,GY= Displays coordinates past the Graphic area.
;												negative coordinates to the up and to the left side
;$h_GameGui                						Game Window/gui handle
;$guiQuit										Used by the Delay Function to end the app.
;$h_User32DLL									Handle to the User32.dll (opened with Initialize, and closed with CleanUp functions)
#include <AutoItConstants.au3>
#include <GDIPlus.au3>
#include <GUIConstantsEx.au3>
#include <Timers.au3>
#include <Misc.au3>
#include <StructureConstants.au3>
#include <WindowsConstants.au3>
#include <WinAPISysInternals.au3>
#include <WinAPISysWin.au3>


Opt("GUICloseOnESC",0)

Global $a_txt_info[7]						;Array which holds the font settings
Global $guiQuit = 0							;check if x is pressed and set to 1 to quit (used by the delay function)
Global $v_GuiInitializeLock = 0
Global $gdip_backbuffer, $gdip_bitmap		;Backbuffer and frontbuffer handles
Global $gdip_GraphicX, $gdip_GraphicY		;Graphics Width and Height
Global $mouseX, $mouseY, $mouseGX, $mouseGY ;Mouse Coordinates x,y=Window based GX,GY= Uses the coordinate outside window
Global $h_GameGui                ;Game gui handle
Global $h_User32DLL            ;Keypress DLL (User32) handle

Func CleanUp()
	If $v_GuiInitializeLock = 1 Then
		; Clean up resources
		_GDIPlus_GraphicsDispose($gdip_backbuffer) ;Free the backbuffer
		_GDIPlus_BrushDispose($a_txt_info[1])
		_GDIPlus_GraphicsDispose($a_txt_info[0])
		_GDIPlus_Shutdown()
		DllClose($h_User32DLL)
	Else
		MsgBox(270384, "Cleanup Error", "Game GDI init was either not started" & @CRLF & "or allready cleaned up.", 10)
	EndIf
EndFunc   ;==>CleanUp

Func Initialize($x = 400, $y = 300, $gamename = "GDI+ test, click on x to continue")
	If $v_GuiInitializeLock = 0 Then
		If $x <= 0 Then $x = 400
		If $y <= 0 Then $y = 300
		$gdip_GraphicX = $x-1
		$gdip_GraphicY = $y-1

		$h_GameGui = _GUICreate($gamename, $gdip_GraphicX, $gdip_GraphicY)  ; Create GUI
		;GUISetState(@SW_SHOW)
		$h_User32DLL = DllOpen("user32.dll")
		_GDIPlus_Startup()                                            ;Startup GDI
		$a_txt_info[0] = _GDIPlus_GraphicsCreateFromHWND($h_GameGui)        ;Create a Graphic object using the gui window
		Color("000000")                                                ;Setup a default Brush/drawing color
		$gdip_bitmap = _GDIPlus_BitmapCreateFromGraphics($gdip_GraphicX, $gdip_GraphicY, $a_txt_info[0])    ;Create bitmap object
		$gdip_backbuffer = _GDIPlus_ImageGetGraphicsContext($gdip_bitmap)     ;Create backbuffer
		$v_GuiInitializeLock = 1
	Else
		MsgBox(270384, "Initialize Error", "Game GDI was allready initialized", 10)
	EndIf

EndFunc   ;==>Initialize

Func CLS($Rgb)
	_GDIPlus_GraphicsClear($gdip_backbuffer, "0xFF" & $Rgb)
EndFunc   ;==>CLS

Func Color($cstr = "000000", $cAlpha = "FF")
	_GDIPlus_BrushDispose($a_txt_info[1])
	$a_txt_info[1] = _GDIPlus_BrushCreateSolid("0x" & $cAlpha & $cstr)
EndFunc   ;==>Color

Func Flip()                ;Draw the backbuffer to the screen, in case you want to use your own delay/sleep functions
	_GDIPlus_GraphicsDrawImageRect($a_txt_info[0], $gdip_bitmap, 0, 0, $gdip_GraphicX, $gdip_GraphicY) ;Draw the backbuffer to the screen
EndFunc   ;==>Flip

Func _WinSetClientPos($sTitle, $sText, $w, $h, $l = Default, $t = Default)
	;Function code example from mat : https://www.autoitscript.com/forum/topic/154224-_winsetclientpos/
    Local Const $iDefStyle = BitOR($WS_MINIMIZEBOX, $WS_CAPTION, $WS_POPUP, $WS_SYSMENU)
    Local Const $iDefExStyle = $WS_EX_WINDOWEDGE

    Local $hWnd = WinGetHandle($sTitle, $sText)
    If @error Then Return SetError(1, 0, 0)

    Local $x = $l, $y = $t

    If IsKeyword($l) = $KEYWORD_DEFAULT Then $x = 100
    If IsKeyword($t) = $KEYWORD_DEFAULT Then $y = 100

    Local $iStyle = _WinAPI_GetWindowLong($hWnd, $GWL_STYLE)
    Local $iExStyle = _WinAPI_GetWindowLong($hWnd, $GWL_EXSTYLE)

    If $iStyle = -1 Then $iStyle = $iDefStyle
    If $iExStyle = -1 Then $iExStyle = $iDefExStyle

    Local $rect = DllStructCreate($tagRECT)
    DllStructSetData($rect, "left", $x)
    DllStructSetData($rect, "right", $x + $w)
    DllStructSetData($rect, "top", $y)
    DllStructSetData($rect, "bottom", $y + $h)

    If Not BitAND($iStyle, BitOR(BitAND($WS_CAPTION, BitNOT($WS_BORDER)), $WS_POPUP)) Then
        _WinAPI_AdjustWindowRectEx($rect, BitOR($iStyle, $WS_CAPTION), $iExStyle, False)
    Else
        _WinAPI_AdjustWindowRectEx($rect, $iStyle, $iExStyle, False)
    EndIf

    $w = DllStructGetData($rect, "right") - DllStructGetData($rect, "left")
    $h = DllStructGetData($rect, "bottom") - DllStructGetData($rect, "top")

    If BitAND($iStyle, $WS_VSCROLL) Then
        $w += _WinAPI_GetSystemMetrics($SM_CXVSCROLL)
    EndIf

    If BitAND($iStyle, $WS_HSCROLL) Then
        $h += _WinAPI_GetSystemMetrics($SM_CYHSCROLL)
    EndIf

    If IsKeyword($l) = $KEYWORD_DEFAULT Then
        $x = (@DesktopWidth - $w) / 2
    Else
        $x = DllStructGetData($rect, "left")
    EndIf

    If IsKeyword($t) = $KEYWORD_DEFAULT Then
        $y = (@DesktopHeight - $h) / 2
    Else
        $y = DllStructGetData($rect, "top")
    EndIf

    Return WinMove($hWnd, "", $x, $y, $w, $h)
Endfunc   ;==>_WinSetClientPos

Func _GUICreate($sTitle, $w, $h, $x = -1, $y = -1, $iStyle = -1, $iExStyle = -1, $hParent = 0)
	;Function Code:mat (Autoit-Forum)
    Local $hRet = GUICreate($sTitle, $w, $h, $x, $y, $iStyle, $iExStyle, $hParent)

    If $x = -1 Then $x = Default
    If $y = -1 Then $y = Default

    _WinSetClientPos($hRet, "", $w, $h, $x, $y)

    Return $hRet
EndFunc   ;==>_GUICreate
